"""DSPSim from pyproject.toml"""

from pathlib import Path
from dataclasses import dataclass, field
import tomllib
import glob
from typing import TypeAlias, Literal
import json
from tempfile import TemporaryDirectory
import shutil
from functools import cache
import numpy as np
from typing import Any
import numpy.typing as npt

# from dspsim import verilator

# from .util import cmake_dir
ParameterValue: TypeAlias = npt.NDArray


@cache
def _uint_width(width: int) -> int:
    if width <= 8:
        return 8
    elif width <= 16:
        return 16
    elif width <= 32:
        return 32
    elif width <= 64:
        return 64


_uint_str = {8: "uint8_t", 16: "uint16_t", 32: "uint32_t", 64: "uint64_t"}


def _vvalue_str(value: ParameterValue):
    if value.shape:
        header = "'{"
        tail = "}"
        inits = ", ".join([_vvalue_str(p) for p in value])
        return f"{header}{inits}{tail}"

    if value.dtype.kind == "U":
        return f'"{value}"'

    return str(value)


def _ctype_str(value: ParameterValue):
    """"""
    if value.shape:
        header = "std::array<"
        tail = f", {len(value)}>"
        # inits = "".join([_cpp_val_type(p) for p in value])
        return f"{header}{_ctype_str(value[0])}{tail}"
    if value.dtype.kind == "U":
        return f"std::string"
    elif value.dtype.kind == "i":
        return "int"
    elif value.dtype.kind == "f":
        return "float"


def _cvalue_str(value: ParameterValue):
    """"""
    if value.shape:
        header = "{" + "{"
        tail = "}" + "}"
        inits = ", ".join([_cvalue_str(p) for p in value])
        return f"{header}{inits}{tail}"
    if value.dtype.kind == "U":
        return f'"{value}"'
    else:
        return str(value)


@dataclass
class Parameter:
    name: str
    value: ParameterValue

    @property
    def ctype(self) -> str:
        """C++ type as a string."""
        return _ctype_str(self.value)

    @property
    def cvalue(self) -> str:
        """C++ initializer for value. As a string."""
        return _cvalue_str(self.value)

    @property
    def vtype(self) -> str:
        """Verilog type as a string."""
        _types = {"i": "int", "f": "float", "U": "string"}
        return _types[self.value.dtype.kind]

    @property
    def vvalue(self) -> str:
        """"""
        return _vvalue_str(self.value)

    @property
    def sv_mod_def(self) -> str:
        """"""
        return f"parameter {self.vtype} {self.name} = {self.vvalue}"

    @property
    def vm_def(self) -> str:
        """"""
        return f"static const {self.ctype} {self.name} = {self.cvalue}"

    def info(self) -> str:
        """Return main information about the parameter."""
        return f"Parameter(name={self.name}, value={self.value}, dtype={self.value.dtype}, shape={self.value.shape})"


def _vm_ctor_arg_str(width: int, shape: tuple) -> str:
    """"""
    if shape:
        return f"std::array<{_vm_ctor_arg_str(width, shape[1:])}, {shape[0]}>"
    return f"dspsim::Signal<{_uint_str[_uint_width(width)]}>*"


@dataclass
class Port:
    """HDL module port configuration."""

    name: str
    width: int
    direction: Literal["input", "output"]  # input or output
    shape: tuple = ()  # Support scalars or arrays

    @property
    def ctype(self) -> str:
        """"""
        return _uint_str[_uint_width(self.width)]

    @property
    def cvalue(self) -> str:
        """"""

    @property
    def vm_port_decl(self) -> str:
        """"""
        port_dir = "Input" if self.direction.startswith("i") else "Output"
        decl = f"dspsim::{port_dir}<{self.ctype}"
        for s in self.shape:
            decl += f"[{s}]"
        return f"{decl}> _{self.name}"

    @property
    def vm_ctor_arg(self) -> str:
        """"""
        if self.shape:
            return f"{_vm_ctor_arg_str(self.width, self.shape)} &{self.name}"
        else:
            return f"dspsim::Signal<{self.ctype}> &{self.name}"

    @property
    def sv_range(self) -> str:
        if self.width <= 1:
            return ""
        return f"[{self.width-1}:0]"

    @property
    def sv_mod_def(self) -> str:
        direction_fmt = "input " if self.direction.startswith("i") else "output"
        width_fmt = f" {self.sv_range}" if self.width > 1 else ""
        unpacked_fmt = "".join([f"[{s}]" for s in self.shape])

        return f"{direction_fmt} logic{width_fmt} {self.name}{unpacked_fmt}"


from dataclass_wizard import JSONSerializable


@dataclass
class ModuleConfig(JSONSerializable):
    """
    The basic information about an hdl source module.
    This can be read from a verilog file, a verilated module,
    Or it can be used from python to create a verilated module or
    a framework model.

    This interface should be compatible with verilog, system verilog, vhdl, etc.
    And other simulators besides verilator could be supported.
    """

    # Parsed and formatted information
    name: str  # This should always be the same as source.stem ?
    source: Path
    parameters: dict[str, Parameter]  # Parsed from type table
    ports: dict[str, Port]  #
    trace: Literal[None, "vcd", "fst", "notrace"] = None
    include_dirs: list[Path] = field(default_factory=list)
    verilator_args: list[str] = field(default_factory=list)

    @classmethod
    def from_verilator(
        cls,
        source: Path,
        parameters: dict[str, Parameter] = {},
        trace: Literal[None, "vcd", "fst", "notrace"] = None,
        include_dirs: list[Path] = [],
        verilator_args: list[str] = [],
    ):
        """"""
        from dspsim.verilator import VModelTree

        tree = VModelTree.generate(
            source,
            parameters=parameters,
            include_dirs=include_dirs,
            verilator_args=verilator_args.copy(),
        )
        # raise Exception(source)
        obj = cls(
            name=tree.name,
            source=source,
            parameters=tree.parameters,
            ports=tree.ports,
            trace=trace,
            include_dirs=include_dirs,
            verilator_args=verilator_args,
        )
        return obj

    def port_info(self) -> str:
        dictinfo = {k: vars(v) for k, v in self.ports.items()}
        return str(dictinfo)
        # return self.to_json()



"""Test reading the pyproject.toml config"""

from pathlib import Path
from dspsim.generate import Config
import pytest


def test_read_tool_config():
    pyproject_path = Path("pyproject.toml")
    config = Config.from_pyproject(pyproject_path)



from dspsim.framework import Signal8
from dspsim.axis import Axis

from dspsim._library import FifoAsync as _FifoAsync, FifoSync as _FifoSync
from dspsim._library import FifoAsync8 as _FifoAsync8


class FifoAsync(_FifoAsync):
    @classmethod
    def init_bus(
        cls,
        clka: Signal8,
        rsta: Signal8,
        s_axis: Axis,
        clkb: Signal8,
        rstb: Signal8,
        m_axis: Axis,
    ):
        return cls(
            clka,
            rsta,
            s_axis.tdata,
            s_axis.tvalid,
            s_axis.tready,
            s_axis.tid,
            clkb,
            rstb,
            m_axis.tdata,
            m_axis.tvalid,
            m_axis.tready,
            m_axis.tid,
        )


class FifoAsync8(_FifoAsync8):
    @classmethod
    def init_bus(
        cls,
        clka: Signal8,
        rsta: Signal8,
        s_axis: Axis,
        clkb: Signal8,
        rstb: Signal8,
        m_axis: Axis,
    ):
        return cls(
            clka,
            rsta,
            s_axis.tdata,
            s_axis.tvalid,
            s_axis.tready,
            s_axis.tid,
            clkb,
            rstb,
            m_axis.tdata,
            m_axis.tvalid,
            m_axis.tready,
            m_axis.tid,
        )


class FifoSync(_FifoSync):
    @classmethod
    def init_bus(cls, clk: Signal8, rst: Signal8, s_axis: Axis, m_axis: Axis):
        """"""
        return cls(
            clk=clk,
            rst=rst,
            s_axis_tdata=s_axis.tdata,
            s_axis_tvalid=s_axis.tvalid,
            s_axis_tready=s_axis.tready,
            s_axis_tid=s_axis.tid,
            m_axis_tdata=m_axis.tdata,
            m_axis_tvalid=m_axis.tvalid,
            m_axis_tready=m_axis.tready,
            m_axis_tid=m_axis.tid,
        )



import logging

logger = logging.getLogger(__name__)

import subprocess
from pathlib import Path
from functools import cache, cached_property
import os
import sys
import json
import shutil
from tempfile import TemporaryDirectory

from dataclasses import dataclass, field
from dspsim.config import Parameter, Port, ModuleConfig
from typing import Literal
import numpy as np


@cache
def verilator_bin() -> Path:
    """"""
    if "VERILATOR_ROOT" in os.environ:
        return Path(os.environ["VERILATOR_ROOT"]) / "bin/verilator"
    else:
        # Assume it's on the path? Search for it?
        return "verilator"


def verilator(args: list[str]):
    """"""
    cmd = [verilator_bin()] + args
    subprocess.run(cmd)


def _int_from_str(s: str) -> int:
    """"""

    def sign_extend(value, bits):
        sign_bit = 1 << (bits - 1)
        return (value & (sign_bit - 1)) - (value & sign_bit)

    fmt, val = s.split("h", maxsplit=1)
    # is_signed = "s" in fmt
    nbits = int(fmt.split("'")[0])

    # print(f"Width: {bits}, {val}, {is_signed}")
    result = int(val, base=16)
    return sign_extend(result, nbits)
    # base_str, val = _rest.split("s", maxsplit=1)


def _range_to_int(r: str) -> int:
    """"""
    ul = [int(x) for x in r.replace("[", "").replace("]", "").split(":")]
    return max(ul) - min(ul) + 1


class VModelTree:
    tree_file: Path
    tree_meta_file: Path

    _name: str

    # Raw json content.
    _content: dict[str]
    _meta_content: dict[str]

    # Raw submodules.
    _miscsp: dict[str]  # Miscs?
    _modulesp: dict[str]  # Modules
    _stmtsp: dict[str]  # Statements?
    _typetable: dict[str, dict[str]]  # Typetable

    def __init__(
        self, tree_file: Path, tree_meta_file: Path = None, source: Path = None
    ):
        """Read in json content."""
        if tree_meta_file is None:
            tree_meta_file = tree_file.with_suffix(".meta.json")
        self.tree_file = tree_file
        self.tree_meta_file = tree_meta_file

        with open(self.tree_file) as fp:
            self._content = json.load(fp)
            # raise Exception(self._content)
        # with open(self.tree_meta_file) as fp:
        #     self._meta_content = json.load(fp)

        self._miscsp = self._content["miscsp"][0]
        self._modulesp = self._content["modulesp"][0]
        self._stmtsp: list[dict[str]] = self._modulesp["stmtsp"]
        self._typetable = {t["addr"]: t for t in self._miscsp["typesp"]}

    @classmethod
    def generate(
        cls,
        source: Path,
        parameters: dict[str, Parameter] = {},
        include_dirs: list[Path] = [],
        verilator_args: list[str] = [],
        output_file: Path = None,
    ):
        """"""

        with TemporaryDirectory() as _tmpdir:
            tmpdir = Path(_tmpdir)
            if len(parameters):
                # We have to check len of parameters before calling, otherwise we recurse endlessly.
                default_tree = VModelTree.generate(
                    source, include_dirs=include_dirs, verilator_args=verilator_args
                )
                # Only allow overriding valid parameters. Throw error if invalid?
                param_args = [
                    f"-G{k}={v.vvalue}"
                    for k, v in parameters.items()
                    if k in default_tree.parameters
                ]
                verilator_args.extend(param_args)

            verilator_args.extend(["--quiet", "--json-only", "--Mdir", tmpdir])
            inc_dirs = [f"-I{d.absolute()}" for d in include_dirs]
            verilator_args.extend(inc_dirs)
            verilator([source] + verilator_args)

            tree_file = tmpdir / f"V{source.stem}.tree.json"
            tree_meta_file = tree_file.with_suffix(".meta.json")

            # Copy file if requested.
            if output_file:
                shutil.copy(tree_file, output_file)

            obj = cls(tree_file, tree_meta_file)
            return obj

    def to_config(self, cls: type[ModuleConfig]) -> ModuleConfig:
        """"""
        return cls(self.name, self.source, self.parameters.copy(), self.ports.copy())

    # @cached_property
    # def source(self) -> Path:
    #     """"""
    #     for f in self._meta_content["files"].values():
    #         filename = Path(f["filename"])
    #         if filename.exists() and filename.suffix in [".v", ".sv"]:
    #             return Path(f["realpath"])

    @cached_property
    def name(self) -> str:
        """"""
        return self._modulesp["name"]

    @cached_property
    def parameters(self) -> dict[str, Parameter]:
        """"""
        _type_names = {
            "logic": _int_from_str,
            "int": _int_from_str,
            "real": float,
            "string": lambda x: str(x)
            .replace("\\", "")
            .replace('"', "")
            .replace("'", ""),
        }

        def _parse_value(valuep):
            """Parse parameter value using typetable."""
            return _type_names[self._typetable[valuep["dtypep"]]["name"]](
                valuep["name"]
            )

        def _parse_param_value(valuep):
            """Recursively read in parameter definition to build multi-dimensional arrays."""
            if "initsp" in valuep:
                return [_parse_param_value(i["valuep"][0]) for i in valuep["initsp"]]

            return _parse_value(valuep)

        # Read in parameters and parse the value.
        return {
            entry["name"]: Parameter(
                entry["name"], np.array(_parse_param_value(entry["valuep"][0]))
            )
            for entry in self._stmtsp
            if entry.get("isGParam", False)
        }

    @cached_property
    def ports(self) -> dict[str, Port]:
        """"""

        def _parse_port_shape(dtypep):
            """"""
            table_type = self._typetable[dtypep]

            if table_type["name"] == "":
                decl_range = _range_to_int(table_type.get("declRange", "[0:0]"))

                ref_type = table_type["refDTypep"]
                ref_declrange = _range_to_int(
                    self._typetable[ref_type].get("declRange", "[0:0]")
                )
                if ref_declrange == 1:
                    return (decl_range,)
                else:
                    return (decl_range, _parse_port_shape(ref_type))

            return ()

        def _parse_port_width(dtypep):
            """Recursively find a ports type. Recursion needed for arrays."""
            table_type = self._typetable[dtypep]
            if table_type["name"] == "":
                return _parse_port_width(table_type["refDTypep"])
            return _range_to_int(table_type.get("range", "[0:0]"))

        def _flatten(data):
            """Flatten a tuple."""
            result = []
            for item in data:
                if isinstance(item, tuple):
                    result.extend(_flatten(item))
                else:
                    result.append(item)
            return tuple(result)

        def _parse_port_entry(entry: dict[str]):
            """"""
            name = entry["name"]
            direction = str(entry["direction"]).lower()
            # dtypep = typetable[entry["dtypep"]]
            dtypep = entry["dtypep"]

            width = _parse_port_width(dtypep)
            shape = _flatten(_parse_port_shape(dtypep))
            return Port(name=name, width=width, direction=direction, shape=shape)

        return {
            entry["name"]: _parse_port_entry(entry)
            for entry in self._stmtsp
            if entry.get("varType", "Not") == "PORT"
        }


# def read_tree(tree_file: Path) -> dict:
#     """Read"""
#     with open(tree_file, "r") as fp:
#         return json.load(fp)


# def write_tree(tree_file: Path, content: dict):
#     """"""
#     with open(tree_file, "w") as fp:
#         json.dump(content, fp)


# def generate_tree(
#     source: Path, output_file: Path = None, *, verilator_args: list[str] = []
# ) -> dict:
#     """"""
#     args = [source, "--json-only"]
#     args.extend(verilator_args)
#     with TemporaryDirectory() as _tmpdir:
#         tmpdir = Path(_tmpdir)

#         args.extend(["--Mdir", tmpdir])
#         verilator(args)

#         tmp_file = tmpdir / f"V{source.stem}.tree.json"

#         content = read_tree(tmp_file)
#         # Copy file if requested.
#         if output_file:
#             shutil.copy(tmp_file, output_file)

#     return content



"""Tests to check if the package was built properly and works."""

from dspsim.framework import Context, Clock, signal, dff, SignalT, Signal8, Model
from dspsim.axis import Axis, AxisTx, AxisRx
from dspsim.library import Skid
import numpy as np


def test_skid_basic():
    with Context(1e-9, 1e-9) as context:
        clk = Clock(10e-9)
        rst = dff(clk, 1)

        b0 = Axis(width=Skid.DW)
        b1 = Axis(width=Skid.DW)

        skid = Skid(clk, rst, *b0, *b1)

        axis_tx = AxisTx(clk, rst, b0)
        axis_rx = AxisRx(clk, rst, b1)

        context.elaborate()

        rst.d = 1
        context.advance(100)
        rst.d = 0
        context.advance(100)

        tx_data = list(range(1, 42))
        axis_tx.write(tx_data)
        context.advance(100)
        axis_rx.tready = 1
        context.advance(1000)

        rx_data = axis_rx.read()
        assert np.all(rx_data == tx_data)



"""
Generate 
"""

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from dataclasses import dataclass, field, fields, replace
from pathlib import Path
import argparse
from typing import Callable
import sys
import os
import tomllib
import json
import shutil
import glob
import numpy as np
import subprocess
from tempfile import TemporaryDirectory
from typing import Literal, TypeAlias

from dataclass_wizard import TOMLWizard, JSONWizard
from functools import cache

from dspsim.config import Parameter, Port, ModuleConfig
from dspsim import util


def _get_abs_path(source: Path, pyproject_path: Path) -> Path:
    if Path(source).is_absolute():
        # An absolute path is given, do nothing.
        return Path(source)
    else:
        # Add relative to pyproject.toml parent if a relative path was given.
        return Path(pyproject_path.parent / source).absolute()


def _find_source(
    source: Path,
    global_sources: list[Path],
    include_dirs: list[Path],
    pyproject_path: Path,
) -> Path:
    """
    If a source is not an absolute path, find it using the configuration.
    Search Order:
    - check if abs.
    - check in global sources list.
    - check in include_dirs (including dspsim/hdl)
    - check relative to pyproject.toml directory.
    """

    if source.is_absolute():
        return source

    # Check global sources
    for s in global_sources:
        if source.name == s.name:
            return s
    # Check global sources stem name
    for s in global_sources:
        if source.stem == s.stem:
            return s
    # Check include_dirs
    for idir in include_dirs:
        if Path(idir / source).exists():
            return Path(idir / source)
    # Check include_dirs stem name?
    for idir in include_dirs:
        for src in idir.iterdir():
            if source.stem == src.stem:
                return src
    # Check relative to pyproject.toml
    if Path(pyproject_path.parent / source).exists():
        return Path(pyproject_path.parent / source)
    # Check stem relative to pyproject.toml
    for src in pyproject_path.parent.iterdir():
        if source.stem == src.stem:
            return src


@dataclass
class Config(JSONWizard):
    # Build config
    library_type: Literal["static", "shared"] = "static"

    # Global parameters
    parameters: dict[str, Parameter] = field(default_factory=dict)

    # Global include dirs
    include_dirs: list[Path] = field(
        default_factory=lambda: [Path(__file__).parent / "hdl"]
    )

    # Sources to use. Will build modules for all sources with default/global settings.
    sources: list[Path] = field(default_factory=list)

    # Tracing. Options vcd, fst, None
    trace: Literal[None, "vcd", "fst", "notrace"] = None

    verilator_args: list = field(default_factory=list)

    # Module configurations. Specify different configurations, non-default options, or overrides.
    models: dict[str, ModuleConfig] = field(default_factory=dict)

    @classmethod
    def from_pyproject(cls, pyproject_path: Path = Path("pyproject.toml")):
        """Read in the complete configuration for all models."""
        pyproject_path = pyproject_path.absolute()

        with open(pyproject_path, "rb") as fp:
            pyproject = tomllib.load(fp)
        dspsim_tool_config: dict = pyproject["tool"]["dspsim"]

        library_type = dspsim_tool_config.get("library_type", "static")
        global_parameters = {
            k: Parameter(k, np.array(v))
            for k, v in dspsim_tool_config.get("parameters", {}).items()
        }
        global_includes = [
            _get_abs_path(Path(p), pyproject_path)
            for p in dspsim_tool_config.get("include_dirs", [])
        ]
        global_includes.append(util.hdl_dir())

        global_trace = dspsim_tool_config.get("trace", None)
        global_vargs = dspsim_tool_config.get("verilator_args", [])

        # Find all default sources
        global_sources: set[Path] = set()
        for source in dspsim_tool_config.get("sources", []):
            # Glob every source, and add all options to the set.
            # fs.append(source)
            abp = _get_abs_path(
                Path(source),
                pyproject_path,
            )
            for filename in glob.glob(str(abp)):
                # fs.append(filename)
                # Add to the set.
                found = _find_source(
                    Path(filename),
                    global_sources,
                    global_includes,
                    pyproject_path,
                )
                global_sources.add(found)
        tool_models: dict = dspsim_tool_config.get("models", {})
        default_module_sources = {source.stem: source for source in global_sources}
        extra_module_sources = {
            model_name: _get_abs_path(
                _find_source(
                    Path(model.get("source", model_name)),
                    global_sources,
                    global_includes + model.get("include_dirs", []),
                    pyproject_path,
                ),
                pyproject_path,
            )
            for model_name, model in tool_models.items()
        }
        all_module_sources = default_module_sources | extra_module_sources
        all_modules: dict[str, ModuleConfig] = {}
        _errors = {}
        for name, _source in all_module_sources.items():
            model = tool_models.get(name, {}).copy()
            _param_cfg = model.get("parameters", {})
            model_parameters = {
                k: Parameter(k, np.array(v)) for k, v in _param_cfg.items()
            }
            model_includes = [
                _get_abs_path(i, pyproject_path) for i in model.get("include_dirs", [])
            ]
            print(f"Name: {name}, Source: {_source}")
            _errors[name] = _source
            # if name == "FifoAsync":
            #     _errors.append(1)
            all_modules[name] = ModuleConfig.from_verilator(
                _source,
                parameters=global_parameters | model_parameters,
                trace=model.get("trace", global_trace),
                include_dirs=global_includes + model_includes,
                verilator_args=global_vargs + model.get("verilator_args", []),
            )
            all_modules[name].name = name
        # raise Exception(_errors)
        return cls(
            library_type=library_type,
            parameters=global_parameters,
            include_dirs=global_includes,
            sources=global_sources,
            trace=global_trace,
            verilator_args=global_vargs,
            models=all_modules,
        )


def run_generate_model(pyproject_path: Path, json_tool_cfg: Path, output_dir: Path):
    """"""
    logger.debug(
        f"run_generate_model. pyproject={pyproject_path}, tool_cfg={json_tool_cfg}, output_dir={output_dir}"
    )
    from dspsim.util import render_template

    config = Config.from_pyproject(pyproject_path)

    lib_odir = output_dir
    os.makedirs(output_dir, exist_ok=True)
    with open(lib_odir / f"{lib_odir.stem}.cpp", "w") as fp:
        fp.write(
            render_template(
                "module_library.cpp.jinja",
                models=config.models,
                libname=f"{lib_odir.stem}",
            )
        )
    for name, model in config.models.items():
        gen_file = output_dir / f"{name}.dir/{name}.h"
        os.makedirs(gen_file.parent.absolute(), exist_ok=True)
        with open(gen_file, "w") as fp:
            # raise Exception(model.port_info())
            fp.write(render_template("model.cpp.jinja", model=model, trace=model.trace))

    from .config import _vvalue_str

    # Convert valid parameters to str values. Arrays are not allowed over command line or cmake.
    for model in config.models.values():
        model.parameters = {
            k: _vvalue_str(v.value)
            for k, v in model.parameters.items()
            if not v.value.shape
        }
    with open(json_tool_cfg, "w") as fp:
        fp.write(config.to_json(indent=4))


@dataclass
class ArgsGenerate:
    pyproject: Path
    tool_cfg: Path
    output_dir: Path
    func: Callable = None

    @classmethod
    def create_parser(cls, subparser: argparse.ArgumentParser = None):
        """"""
        help_str = "Generate code from templates."
        if subparser is None:
            parser = argparse.ArgumentParser("dspsim-generate")
        else:
            parser = subparser.add_parser("generate", help=help_str)
        parser.add_argument(
            "--pyproject",
            type=Path,
            default=Path("pyproject.toml"),
            help="pyproject.toml config.",
        )
        parser.add_argument(
            "--tool-cfg",
            type=Path,
            help="dspsim json tool config.",
        )
        parser.add_argument(
            "--output-dir", type=Path, help="Output dir for generated models."
        )
        parser.set_defaults(func=ArgsGenerate.parse_args)
        return parser

    @classmethod
    def parse_args(cls, cli_args: list[str] = None):
        parser = cls.create_parser()
        cargs = parser.parse_args(cli_args)
        return cls(**vars(cargs))


def main(cli_args: list[str] = None):
    """"""
    args = ArgsGenerate.parse_args(cli_args)
    # create console handler and set level to debug
    log_file = args.output_dir / "generate.log"
    args.output_dir.mkdir(exist_ok=True)
    ch = logging.FileHandler(log_file, mode="w")
    ch.setLevel(logging.DEBUG)

    # create formatter
    formatter = logging.Formatter("%(levelname)s - %(message)s")

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logging.getLogger(__name__).addHandler(ch)

    logger.debug(f"generate.main args: {args}")

    run_generate_model(args.pyproject, args.tool_cfg, args.output_dir)


if __name__ == "__main__":
    exit(main())



# from dspsim._framework import *

# from dspsim._framework import get_context_factory
from dspsim._framework import Context, global_context
from dspsim._framework import Model, Clock
from dspsim._framework import Signal8, Signal16, Signal32, Signal64
from dspsim._framework import Dff8, Dff16, Dff32, Dff64

import contextlib as _contextlib
import functools as _functools
from dspsim import util as _util
from typing import TypeVar, Literal

SignalT = Signal8 | Signal16 | Signal32 | Signal64
DffT = Dff8 | Dff16 | Dff32 | Dff64

ModelT = TypeVar("ModelT", bound=Model)


def _sclass(width: int) -> type[SignalT]:
    """"""
    _types = {8: Signal8, 16: Signal16, 32: Signal32, 64: Signal64}
    return _types[_util.uint_width(width)]


def _dffclass(width: int) -> type[DffT]:
    """"""
    _types = {8: Dff8, 16: Dff16, 32: Dff32, 64: Dff64}
    return _types[_util.uint_width(width)]


def _signal(initial: int = 0, *, width: int = 1) -> SignalT:
    """Create a signal of the correct stdint type based on the bitwidth."""
    return _sclass(width)(initial)


def signal(
    initial: int = 0, *, width: int = 1, shape: tuple = ()
) -> SignalT | list[SignalT]:
    """
    Create a signal or signal array with the appropriate shape.
    This builds up the list recursively based on the shape.
    """
    if len(shape):
        return [signal(initial, width=width, shape=shape[1:]) for i in range(shape[0])]
    return _signal(initial, width=width)


def _dff(clk: Signal8, initial: int = 0, *, width: int = 1) -> DffT:
    """Create a signal of the correct stdint type based on the bitwidth."""
    return _dffclass(width)(clk, initial)


def dff(
    clk: Signal8, initial: int = 0, *, width: int = 1, shape: tuple = ()
) -> DffT | list[DffT]:
    """"""
    if len(shape):
        return [
            dff(clk, initial, width=width, shape=shape[1:]) for _ in range(shape[0])
        ]
    return _dff(clk, initial, width=width)


@_contextlib.contextmanager
def enter_context(time_unit: float = 1e-9, time_precision: float = 1e-9):
    context = Context(time_unit, time_precision)
    try:
        yield context
    finally:
        context.clear()


def runner(time_unit: float = 1e-9, time_precision: float = 1e-9):
    """"""

    def runner_deco(func):
        @_functools.wraps(func)
        def wrapped():
            context = Context(time_unit, time_precision)
            result = func(context)
            context.clear()
            return result

        return wrapped

    return runner_deco


from dspsim.config import Port as _Port


def port_info(model: Model) -> dict[str, _Port]:
    """"""
    import ast

    linfo = ast.literal_eval(model.port_info)
    return {k: _Port(**v) for k, v in linfo.items()}



from dspsim.framework import Model, Signal8, SignalT, signal, port_info

from dspsim._framework import AxisTx8, AxisTx16, AxisTx32, AxisTx64
from dspsim._framework import AxisRx8, AxisRx16, AxisRx32, AxisRx64

import numpy as _np
from numpy.typing import ArrayLike as _ArrayLike
from typing import TypeVar, Literal

TIDW = 8


class Axis:
    tdata: SignalT
    tvalid: Signal8
    tready: Signal8
    tid: Signal8 = None
    tlast: Signal8 = None

    _width: int

    def __init__(self, *, width: int, tid: bool = False, tlast: bool = False):
        self._width = width
        self.tdata = signal(width=width)
        self.tvalid = signal()
        self.tready = signal()
        if tid:
            self.tid = signal(width=TIDW)
        if tlast:
            self.tlast = signal()

    def __str__(self) -> str:
        return f"Axis(width={self.width}, tid={self.tid}, tlast={self.tlast})"

    def __iter__(self):
        return iter((self.tdata, self.tvalid, self.tready))

    @property
    def width(self):
        return self._width


import itertools


def init_stream_model[
    ModelT
](
    cls: type[ModelT],
    clk: Signal8,
    rst: Signal8,
    s_axis: Axis,
    m_axis: Axis,
    **extra,
) -> ModelT:
    """
    Init a model that contains a stream input and output using
    """
    from dspsim.config import Port

    args = dict(
        clk=clk,
        rst=rst,
        s_axis_tdata=s_axis.tdata,
        s_axis_tvalid=s_axis.tvalid,
        s_axis_tready=s_axis.tready,
        m_axis_tdata=m_axis.tdata,
        m_axis_tvalid=m_axis.tvalid,
        m_axis_tready=m_axis.tready,
    )

    portinfo = port_info(cls)

    # Connect an empty signal if the bus doesn't have it.
    # If the model doesn't have it, don't connect it.
    if "s_axis_tid" in portinfo:
        args["s_axis_tid"] = s_axis.tid if s_axis.tid else signal(width=TIDW)
    if "s_axis_tlast" in portinfo:
        args["s_axis_tlast"] = s_axis.tlast if s_axis.tlast else signal(width=1)

    if "m_axis_tid" in portinfo:
        args["m_axis_tid"] = m_axis.tid if m_axis.tid else signal(width=TIDW)
    if "m_axis_tlast" in portinfo:
        args["m_axis_tlast"] = m_axis.tlast if m_axis.tlast else signal(width=1)

    return cls(
        **args,
        **extra,
    )


AxisTxT = AxisTx8 | AxisTx16 | AxisTx32 | AxisTx64
AxisRxT = AxisRx8 | AxisRx16 | AxisRx32 | AxisRx64
from dspsim import util


def AxisTx(
    clk: Signal8,
    rst: Signal8,
    m_axis: Axis,
    tid_pattern: list[int] = [0],
    *,
    width: int = 32,
) -> AxisTxT:
    """"""

    _models = {8: AxisTx8, 16: AxisTx16, 32: AxisTx32, 64: AxisTx64}
    cls = _models[util.uint_width(width)]
    return cls(
        clk=clk,
        rst=rst,
        m_axis_tdata=m_axis.tdata,
        m_axis_tvalid=m_axis.tvalid,
        m_axis_tready=m_axis.tready,
        m_axis_tid=m_axis.tid,
        tid_pattern=tid_pattern,
    )


def AxisRx(
    clk: Signal8,
    rst: Signal8,
    s_axis: Axis,
    *,
    width: int = 32,
) -> AxisRxT:
    """"""
    _models = {8: AxisRx8, 16: AxisRx16, 32: AxisRx32, 64: AxisRx64}
    cls = _models[util.uint_width(width)]
    return cls(clk, rst, s_axis.tdata, s_axis.tvalid, s_axis.tready, s_axis.tid)



# from dspsim._library import *
import dspsim._framework
import dspsim.framework
from dspsim.framework import Signal8, SignalT

# Some components have additional wrappers.
from dspsim.axis import Axis
from dspsim.fifo import FifoSync, FifoAsync, FifoAsync8
from dspsim.i2s import I2S, I2SClkGen, I2STx, I2SRx

from dspsim._library import link_context as _link_context

_link_context(dspsim.framework.global_context())

from dspsim._library import Foo, SomeModel, X
from dspsim._library import AsyncSync8, AsyncSync

from dspsim._library import Skid as _Skid
from dspsim._library import Gain as _Gain

# # Sync to the framework context factory.
# from dspsim._library import set_context_factory

# set_context_factory(dspsim.framework.get_context_factory())


class Skid(_Skid):
    """Wrapper to allow connecting axis busses"""

    @classmethod
    def init_bus(cls, clk: Signal8, rst: Signal8, s_axis: Axis, m_axis: Axis):
        """"""
        return cls(
            clk=clk,
            rst=rst,
            s_axis_tdata=s_axis.tdata,
            s_axis_tvalid=s_axis.tvalid,
            s_axis_tready=s_axis.tready,
            m_axis_tdata=m_axis.tdata,
            m_axis_tvalid=m_axis.tvalid,
            m_axis_tready=m_axis.tready,
        )


class Gain(_Gain):
    """Wrapper to allow connecting axis busses"""

    @classmethod
    def init_bus(
        cls, clk: Signal8, rst: Signal8, s_axis: Axis, m_axis: Axis, gain: SignalT
    ):
        """"""
        return cls(
            clk=clk,
            rst=rst,
            s_axis_tdata=s_axis.tdata,
            s_axis_tvalid=s_axis.tvalid,
            s_axis_tready=s_axis.tready,
            m_axis_tdata=m_axis.tdata,
            m_axis_tvalid=m_axis.tvalid,
            m_axis_tready=m_axis.tready,
            gain=gain,
        )



from dspsim.framework import Context, Clock, signal, dff, SignalT, Signal8, Model
from dspsim.axis import Axis, AxisTx, AxisRx
from dspsim.library import Skid
import numpy as np


def test_skid_basic():
    context = Context(1e-9, 1e-9)

    clk = Clock(10e-9)
    rst = dff(clk, 1)

    b0 = Axis(width=Skid.DW)
    b1 = Axis(width=Skid.DW)
    b2 = Axis(width=Skid.DW)
    print(context.print_info())

    skid0 = Skid.init_bus(clk, rst, b0, b1)
    skid1 = Skid.init_bus(clk, rst, b1, b2)

    print(context.print_info())
    axis_tx = AxisTx(clk, rst, b0)
    axis_rx = AxisRx(clk, rst, b2)

    print(context.print_info())

    skid0.trace("traces/skid0.vcd")
    skid1.trace("traces/skid1.vcd")

    context.elaborate()

    context.advance(100)

    rst.d = 0
    context.advance(100)

    tx_data = list(range(1, 6))
    axis_tx.write(tx_data)
    context.advance(100)
    axis_rx.tready = 1
    context.advance(100)

    rx_data = axis_rx.read()

    assert np.all(tx_data == rx_data)



"""Miscellaneous Utilities"""

from pathlib import Path
import functools as _functools


def cmake_dir() -> Path:
    return Path(__file__).parent / "lib/cmake/dspsim"


def include_dir() -> Path:
    return Path(__file__).parent / "include"


def hdl_dir() -> Path:
    return Path(Path(__file__).parent / "hdl").absolute()


def lib_dir() -> Path:
    return Path(__file__).parent / "lib"


import jinja2

_template_env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates")
)


def render_template(template: str, **kwargs) -> str:
    """"""
    return _template_env.get_template(template).render(kwargs)


@_functools.cache
def uint_width(width: int) -> int:
    if width <= 8:
        return 8
    elif width <= 16:
        return 16
    elif width <= 32:
        return 32
    elif width <= 64:
        return 64


def to_fixed(flt: float, q: int) -> int:
    """"""
    return flt * (2**q)


def to_float(fxd: int, q: int) -> float:
    """"""
    return fxd / (2**q)



from dspsim.framework import Context, Clock, signal, dff, SignalT, Signal8, Model
from dspsim.axis import Axis, AxisTx, AxisRx, init_stream_model
from dspsim.library import Gain, FifoSync
import numpy as np
from dspsim.util import to_fixed, to_float


def test_gain_basic():
    with Context(1e-9, 1e-9) as context:
        clk = Clock(10e-9)
        rst = dff(clk, 1)

        b0 = Axis(width=Gain.DW, tid=True)
        b1 = Axis(width=Gain.DW, tid=True)
        b2 = Axis(width=Gain.DW)

        _k = to_fixed(0.1, Gain.COEFQ)
        k = signal(int(_k), width=Gain.COEFW)

        print(context.print_info())

        fifo = FifoSync.init_bus(clk, rst, b0, b1)
        gain = Gain.init_bus(clk, rst, b1, b2, gain=k)
        axis_tx = AxisTx(clk, rst, b0)
        axis_rx = AxisRx(clk, rst, b2)

        print(context.print_info())

        gain.trace("traces/gain.vcd")

        context.elaborate()

        context.advance(100)

        rst.d = 0
        context.advance(100)

        DATAQ = 16
        tx_data = np.linspace(1, 10.0, 10)
        axis_tx.write(tx_data, DATAQ)
        context.advance(100)
        axis_rx.tready = True
        context.advance(200)

        rx_data = to_float(np.array(axis_rx.read()), DATAQ)
        assert np.all(np.isclose(rx_data * 10, tx_data, rtol=0.0001))
